# ai_hydra/constants/DEvent.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from ai_hydra.constants.DHydraTui import DField
from ai_hydra.constants.DHydra import DModule


EVENT_MAP = {
    DField.SIM_LOOP: "⚡ Simulation",
    DField.SNAPSHOT: "📸 Snapshot",
    DField.TUI: "💻 Hydra Client",
    DField.WARNING: "⚠️ Warning",
    DModule.EPSILON_ALGO: "🎲 Epsilon",
}
